from django.urls import include, path
from rest_framework.schemas import get_schema_view

from smoothglue.core.views import get_app_info

schema_view = get_schema_view(
    title="SmoothGlue Core",
    description="SmoothGlue Core API",
    version="1.0.0",
    patterns=[
        path("", include("smoothglue.core.urls")),
    ],
)

urlpatterns = [
    path("app_info/", get_app_info, name="app_info"),
    # path("openapi/", schema_view, name="openapi-schema"),
]



from html import unescape
from typing import Any, Dict

from django.utils.html import escape
from rest_framework import serializers

from smoothglue.core.models import AuditModel, TimeAuditModel


class DefaultSerializer(serializers.ModelSerializer):
    """
    Generic serializer to ensure all values are escaped
    """

    def validate(self, attrs: Dict[str, Any]) -> Dict[str, Any]:
        """
        Validates and sanitizes the input data.

        This method iterates over all attributes in the input data. If an attribute
        is of type string, it escapes any HTML characters to prevent XSS attacks.
        This is particularly useful for ensuring that user input is safely
        processed and stored.

        Parameters:
        attrs (Dict[str, Any]): The input data to validate, as a dictionary
            where keys are the field names and values are the field values.

        Returns:
        Dict[str, Any]: The sanitized input data, with HTML characters in string
            fields escaped.
        """
        for key, value in attrs.items():
            if isinstance(value, str):
                attrs[key] = escape(value)
        return super().validate(attrs)

    def to_representation(self, instance):
        """
        Returned the unescaped value for str typed fields.
        """
        data = super().to_representation(instance)
        for key, value in data.items():
            if isinstance(value, str):
                data[key] = unescape(value)
        return data

    class Meta:
        abstract = True


class TimeAuditSerializer(DefaultSerializer):
    """
    A serializer for models that include time audit fields.

    This serializer extends DefaultSerializer and is specifically designed for
    models that have 'created' and 'last_updated' fields. These fields are
    automatically set to read-only to ensure they are managed by the model itself
    and not modifiable through the serializer.

    Attributes:
        model (Model): A Django model that includes time audit fields.
        fields (list[str]): The fields to be included in the serialization.
        read_only_fields (list[str]): Fields that are read-only.
        abstract (bool): Marks the serializer as abstract.
    """

    class Meta:
        model = TimeAuditModel
        fields = ["created_at", "updated_at"]
        read_only_fields = fields
        abstract = True


class AuditSerializer(DefaultSerializer):
    """
    A serializer for models that include both time and user audit fields.

    This serializer extends DefaultSerializer and is specifically designed for
    models that, in addition to 'created' and 'last_updated' fields, also include
    'created_by' and 'updated_by' fields to track the user responsible for
    creating or updating the model instance. The user information is automatically
    extracted from the request context and applied to the model instance.

    Attributes:
        model (Model): A Django model that includes user and time audit fields.
        fields (list[str]): The fields to be included in the serialization.
        read_only_fields (list[str]): Fields that are read-only.
        abstract (bool): Marks the serializer as abstract.
    """

    class Meta:
        model = AuditModel
        fields = TimeAuditSerializer.Meta.fields + ["created_by", "updated_by"]
        read_only_fields = fields
        abstract = True

    def create(self, validated_data: Dict[str, Any]) -> Any:
        """
        Overrides the create method to add 'created_by' and 'updated_by'
        fields based on the current user before creating a new model instance.

        Parameters:
        validated_data (Dict[str, Any]): The validated data used for model instance creation.

        Returns:
        Any: The newly created model instance.
        """
        user = self.context["request"].user
        if user:
            validated_data["updated_by"] = user
            validated_data["created_by"] = user
        return super().create(validated_data)

    def update(self, instance: Any, validated_data: Dict[str, Any]) -> Any:
        """
        Overrides the update method to set the 'updated_by' field based
        on the current user before updating the model instance.

        Parameters:
        instance (Any): The model instance to update.
        validated_data (Dict[str, Any]): The validated data used for model instance updating.

        Returns:
        Any: The updated model instance.
        """
        user = self.context["request"].user
        if user:
            instance.updated_by = user
        return super().update(instance, validated_data)



import unittest
from unittest.mock import Mock, patch

from django.core.exceptions import ObjectDoesNotExist
from django.db import models
from rest_framework import exceptions, serializers

from smoothglue.core.serializers.dynamic import (
    DynamicModelSerializer,
    NestedObjectModelSerializer,
    handle_many_to_many_objects,
    handle_nested_objects,
)

UPDATED_OBJECT_TEXT = "Updated Object"


class MockUser(models.Model):
    id = models.IntegerField(primary_key=True)
    username = models.CharField(max_length=100)

    class Meta:
        app_label = "test_app"


class MockProfile(models.Model):
    id = models.IntegerField(primary_key=True)
    user = models.OneToOneField(
        MockUser, on_delete=models.CASCADE, related_name="profile"
    )
    bio = models.TextField()

    class Meta:
        app_label = "test_app"


class MockPost(models.Model):
    id = models.IntegerField(primary_key=True)
    author = models.ForeignKey(MockUser, on_delete=models.CASCADE, related_name="posts")
    title = models.CharField(max_length=100)

    class Meta:
        app_label = "test_app"


class MockTag(models.Model):
    id = models.IntegerField(primary_key=True)
    name = models.CharField(max_length=50)
    posts = models.ManyToManyField(MockPost, related_name="tags")

    class Meta:
        app_label = "test_app"


class MockTagSerializer(serializers.ModelSerializer):
    class Meta:
        model = MockTag
        fields = ["id", "name"]


class MockPostSerializer(NestedObjectModelSerializer):
    tags = MockTagSerializer(many=True, required=False)
    author_id = serializers.PrimaryKeyRelatedField(
        queryset=MockUser.objects.all(), source="author", write_only=True
    )

    class Meta:
        model = MockPost
        fields = ["id", "title", "tags", "author_id"]

    parent_id_field = "author_id"
    nested_field_config = {
        "tags": {"serializer": MockTagSerializer, "model": MockTag},
    }


class MockPostNestedSerializer(MockPostSerializer):
    pass


class MockUserSerializer(NestedObjectModelSerializer):
    posts = MockPostNestedSerializer(many=True, required=False)
    profile_bio = serializers.CharField(source="profile.bio", required=False)

    class Meta:
        model = MockUser
        fields = ("id", "username", "posts", "profile_bio")

    # For NestedObjectModelSerializer
    parent_id_field = "user_id"
    nested_field_config = {
        "posts": {"serializer": MockPostSerializer, "model": MockPost},
    }


class TestDynamicModelSerializer(unittest.TestCase):
    def setUp(self):
        self.mock_request = Mock()
        self.mock_request.query_params = {}
        self.context = {"request": self.mock_request}

    @patch("smoothglue.core.field_trie.ModifiedTrie")
    def test_init_no_fields_in_request(self, MockModifiedTrie):
        class MySerializer(DynamicModelSerializer):
            field_a = serializers.CharField()
            field_b = serializers.CharField()

            class Meta:
                model = MockUser  # Dummy model
                fields = ["field_a", "field_b"]

        serializer = MySerializer(context=self.context)
        self.assertIn("field_a", serializer.fields)
        self.assertIn("field_b", serializer.fields)
        MockModifiedTrie.return_value.filter_requested_fields.assert_not_called()

    def test_validate_partial_update_with_id_fetches_instance(self):
        class MySerializer(DynamicModelSerializer):
            class Meta:
                model = MockUser
                fields = ["id", "username"]

        mock_user_instance = Mock(id=1, username="testuser")
        MockUser.objects = Mock()
        MockUser.objects.get = Mock(return_value=mock_user_instance)

        serializer = MySerializer(partial=True, context=self.context)
        self.assertIsNone(serializer.instance)

        attrs = {"id": 1, "username": "newname"}
        validated_attrs = serializer.validate(attrs)

        MockUser.objects.get.assert_called_once_with(id=1)
        self.assertEqual(serializer.instance, mock_user_instance)
        self.assertEqual(validated_attrs, attrs)

    def test_validate_patch_request_with_id_fetches_instance(self):
        class MySerializer(DynamicModelSerializer):
            class Meta:
                model = MockUser
                fields = ["id", "username"]

        mock_user_instance = Mock(id=1, username="testuser")
        MockUser.objects = Mock()
        MockUser.objects.get = Mock(return_value=mock_user_instance)

        self.mock_request.method = "PATCH"
        serializer = MySerializer(context=self.context)
        self.assertIsNone(serializer.instance)

        attrs = {"id": 1, "username": "newname"}
        serializer.validate(attrs)

        MockUser.objects.get.assert_called_once_with(id=1)
        self.assertEqual(serializer.instance, mock_user_instance)


class TestNestedObjectModelSerializer(unittest.TestCase):
    def setUp(self):
        self.mock_request = Mock()
        self.mock_request.query_params = {}
        self.context = {"request": self.mock_request}

        import uuid

        unique_suffix = str(uuid.uuid4())[:8]

        class SimpleParentModel(models.Model):
            id = models.IntegerField(primary_key=True)
            name = models.CharField(max_length=50)

            class Meta:
                app_label = f"test_app_parent_{unique_suffix}"

        self.ParentModel = SimpleParentModel

        class SimpleChildModel(models.Model):
            id = models.IntegerField(primary_key=True)
            parent = models.ForeignKey(
                SimpleParentModel, on_delete=models.CASCADE, related_name="children"
            )
            content = models.CharField(max_length=50)

            class Meta:
                app_label = f"test_app_child_{unique_suffix}"

        self.ChildModel = SimpleChildModel

        class SimpleChildSerializer(NestedObjectModelSerializer):
            class Meta:
                model = SimpleChildModel
                fields = ["id", "content", "parent_id"]

            parent_id_field = "parent_id"

            def get_nested_field_config(self):
                return {}

        class SimpleParentSerializer(NestedObjectModelSerializer):
            children = SimpleChildSerializer(many=True, required=False)

            class Meta:
                model = SimpleParentModel
                fields = ["id", "name", "children"]

            parent_id_field = "some_outer_parent_id"
            nested_field_config = {
                "children": {
                    "serializer": SimpleChildSerializer,
                    "model": SimpleChildModel,
                }
            }

        self.ParentSerializer = SimpleParentSerializer
        self.ChildSerializer = SimpleChildSerializer

        self.ParentModel.objects = Mock()
        self.ChildModel.objects = Mock()

    def test_get_parent_id_field_assertion(self):
        class FaultySerializer(NestedObjectModelSerializer):
            class Meta:
                model = self.ParentModel
                fields = "__all__"

            def get_nested_field_config(self):
                return {"some": "config"}

        with self.assertRaisesRegex(
            AssertionError, "should either include a `parent_id_field`"
        ):
            FaultySerializer()

    def test_get_nested_field_config_assertion(self):
        class FaultySerializer(NestedObjectModelSerializer):
            parent_id_field = "some_field"

            class Meta:
                model = self.ParentModel
                fields = "__all__"

        with self.assertRaisesRegex(
            AssertionError, "should either include a `nested_field_config`"
        ):
            FaultySerializer()

    def test_get_source_field_name(self):
        class MySerializer(NestedObjectModelSerializer):
            renamed_field = serializers.CharField(source="original_field")
            normal_field = serializers.CharField()
            parent_id_field = "dummy"
            nested_field_config = {"dummy": {"serializer": Mock(), "model": Mock()}}

            class Meta:
                model = MockUser
                fields = "__all__"

        serializer = MySerializer()
        self.assertEqual(
            serializer.get_source_field_name("renamed_field"), "original_field"
        )
        self.assertEqual(
            serializer.get_source_field_name("normal_field"), "normal_field"
        )

    def test_get_obj_data(self):
        serializer = self.ParentSerializer()
        validated_data = {
            "name": "Parent Name",
            "children": [
                {"content": "Child 1"},
                {"content": "Child 2"},
            ],
            "extra_field": "should be ignored",
        }
        serializer._declared_fields["children"] = Mock(source="children")

        obj_data = serializer.get_obj_data(
            serializer.nested_field_config, validated_data
        )

        self.assertIn("children", obj_data)
        self.assertNotIn("children", validated_data)
        self.assertIn("name", validated_data)

    @patch("smoothglue.core.serializers.dynamic.handle_nested_objects")
    def test_iterate_nested_data_handles_children(self, mock_handle_nested):
        """
        Tests that iterate_nested_data correctly calls helper functions
        for creating and updating nested objects.
        """
        mock_parent_instance = Mock(spec=self.ParentModel)
        mock_parent_instance._meta = Mock()
        mock_parent_instance._meta.get_field.return_value.many_to_many = False

        new_child_data = {"content": "New Child"}
        updated_child_data = {"id": 1, "content": "Updated Child"}
        obj_data = {"children": [updated_child_data, new_child_data]}

        mock_handle_nested.side_effect = [Mock(id=1), Mock(id=2)]

        serializer = self.ParentSerializer(instance=mock_parent_instance)

        updated_ids = serializer.iterate_nested_data(
            obj_data=obj_data,
            parent_instance=mock_parent_instance,
            model=self.ChildModel,
            field="children",
            serializer=self.ChildSerializer,
            method="update",
            partial=True,
        )

        self.assertEqual(mock_handle_nested.call_count, 2)
        mock_handle_nested.assert_any_call(
            updated_child_data, self.ChildModel, self.ChildSerializer, True
        )
        mock_handle_nested.assert_any_call(
            new_child_data, self.ChildModel, self.ChildSerializer, True
        )

        self.assertEqual(len(updated_ids), 2)
        self.assertIn(1, updated_ids)
        self.assertIn(2, updated_ids)

    def test_update_deletes_children_on_non_partial_update(self):
        """
        Tests that the main `update` method calls the deletion logic
        for a non-partial (PUT) update.
        """
        mock_parent_instance = Mock(spec=self.ParentModel, id=1)
        mock_parent_instance._meta = Mock()
        mock_parent_instance._meta.get_field.return_value.many_to_many = False
        mock_parent_instance._meta.get_field.return_value.one_to_one = False

        mock_qs = Mock()
        self.ChildModel.objects.filter.return_value = mock_qs
        mock_qs.exclude.return_value = mock_qs
        mock_qs.all.return_value = [Mock(name="obj_to_delete")]

        validated_data = {"children": [{"id": 1, "content": "Kept Child"}]}

        with patch.object(DynamicModelSerializer, "update") as mock_super_update:
            serializer = self.ParentSerializer(instance=mock_parent_instance)
            serializer.get_obj_data = Mock(return_value=validated_data)
            serializer.iterate_nested_data = Mock(return_value=[1])

            serializer.update(mock_parent_instance, validated_data)

            # Assert deletion logic was called correctly
            self.ChildModel.objects.filter.assert_called_once_with(
                some_outer_parent_id=1
            )
            mock_qs.exclude.assert_called_once_with(pk__in=[1])
            self.assertEqual(mock_qs.all.return_value[0].delete.call_count, 1)

            mock_super_update.assert_called_once()


class TestHelperFunctions(unittest.TestCase):
    def setUp(self):
        self.MockModel = Mock(spec=models.Model)
        self.MockModel.__name__ = "MockModel"
        self.MockModel.objects = Mock()
        self.MockSerializer = Mock(spec=serializers.ModelSerializer)

    def test_handle_nested_objects_create(self):
        data = {"name": "New Object"}
        mock_serializer_instance = Mock()
        mock_created_object = Mock(id=1)
        mock_serializer_instance.create = Mock(return_value=mock_created_object)
        self.MockSerializer.return_value = mock_serializer_instance

        instance = handle_nested_objects(
            data, self.MockModel, self.MockSerializer, partial=False
        )

        self.MockSerializer.assert_called_once_with()
        mock_serializer_instance.create.assert_called_once_with(data)
        self.assertEqual(instance, mock_created_object)
        self.MockModel.objects.get.assert_not_called()

    def test_handle_nested_objects_update(self):
        data = {"id": 1, "name": UPDATED_OBJECT_TEXT}
        mock_existing_object = Mock(id=1)
        mock_updated_object = Mock(id=1, name="Updated Object From Serializer")
        self.MockModel.objects.get = Mock(return_value=mock_existing_object)

        mock_serializer_instance = Mock()
        mock_serializer_instance.update = Mock(return_value=mock_updated_object)
        self.MockSerializer.return_value = mock_serializer_instance

        instance = handle_nested_objects(
            data, self.MockModel, self.MockSerializer, partial=True
        )

        self.MockModel.objects.get.assert_called_once_with(id=1)
        self.MockSerializer.assert_called_once_with(partial=True)
        mock_serializer_instance.update.assert_called_once_with(
            mock_existing_object, data
        )
        self.assertEqual(instance, mock_updated_object)

    def test_handle_nested_objects_update_not_found(self):
        data = {"id": 99, "name": UPDATED_OBJECT_TEXT}
        self.MockModel.objects.get = Mock(side_effect=ObjectDoesNotExist)

        with self.assertRaisesRegex(
            exceptions.NotFound, "MockModel with id: 99 not found."
        ):
            handle_nested_objects(
                data, self.MockModel, self.MockSerializer, partial=True
            )
        self.MockModel.objects.get.assert_called_once_with(id=99)

    def test_handle_nested_objects_partial_update_missing_id_raises_error(self):
        data = {"name": UPDATED_OBJECT_TEXT}
        with self.assertRaisesRegex(
            serializers.ValidationError, "'id' required for nested update to MockModel."
        ):
            handle_nested_objects(
                data, self.MockModel, self.MockSerializer, partial=True
            )

    def test_handle_many_to_many_objects_create_mode(self):
        mock_child_instance = Mock()
        mock_child_instance.tags = Mock()
        mock_parent_instance = Mock()
        mock_parent_instance.posts = Mock()

        kwargs = {
            "child_instance": mock_child_instance,
            "parent_instance": mock_parent_instance,
            "related_name": "tags",
            "relationship_ids": [1, 2],
            "field": "posts",
            "updated_ids": [mock_child_instance.id],
            "partial": False,
            "method": "create",
            "model": Mock(),
            "serializer": Mock(),
        }

        handle_many_to_many_objects(**kwargs)

        mock_child_instance.tags.set.assert_called_once_with([1, 2])
        mock_parent_instance.posts.add.assert_called_once_with(mock_child_instance)
        mock_parent_instance.posts.set.assert_not_called()  # Not for create

    def test_handle_many_to_many_objects_update_mode_partial(self):
        mock_child_instance = Mock()
        mock_child_instance.tags = Mock()
        mock_parent_instance = Mock()
        mock_parent_instance.posts = Mock()

        kwargs = {
            "child_instance": mock_child_instance,
            "parent_instance": mock_parent_instance,
            "related_name": "tags",
            "relationship_ids": [3, 4],
            "field": "posts",
            "updated_ids": [mock_child_instance.id, 99],
            "partial": True,
            "method": "update",
            "model": Mock(),
            "serializer": Mock(),
        }

        handle_many_to_many_objects(**kwargs)

        mock_child_instance.tags.set.assert_called_once_with([3, 4])
        mock_parent_instance.posts.add.assert_called_once_with(mock_child_instance)
        mock_parent_instance.posts.set.assert_not_called()

    def test_handle_many_to_many_objects_update_mode_not_partial(self):
        mock_child_instance = Mock()
        mock_child_instance.tags = Mock()
        mock_parent_instance = Mock()
        mock_parent_instance.posts = Mock()
        updated_child_ids = [mock_child_instance.id, 99]

        kwargs = {
            "child_instance": mock_child_instance,
            "parent_instance": mock_parent_instance,
            "related_name": "tags",
            "relationship_ids": [5, 6],
            "field": "posts",
            "updated_ids": updated_child_ids,
            "partial": False,
            "method": "update",
            "model": Mock(),
            "serializer": Mock(),
        }

        handle_many_to_many_objects(**kwargs)

        mock_child_instance.tags.set.assert_called_once_with([5, 6])
        mock_parent_instance.posts.add.assert_not_called()
        mock_parent_instance.posts.set.assert_called_once_with(updated_child_ids)



# Generated by Django 5.2.7 on 2025-10-22 20:57

from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("authentication", "0007_alter_organizationcategory_options"),
    ]

    operations = [
        migrations.AlterField(
            model_name="platformorganization",
            name="platform_users",
            field=models.ManyToManyField(
                default=None,
                through="authentication.OrganizationMember",
                through_fields=["platform_organization", "platform_user"],
                to=settings.AUTH_USER_MODEL,
            ),
        ),
        migrations.AlterField(
            model_name="platformuser",
            name="organizations",
            field=models.ManyToManyField(
                default=None,
                related_name="members",
                through="authentication.OrganizationMember",
                through_fields=["platform_user", "platform_organization"],
                to="authentication.platformorganization",
            ),
        ),
    ]



# Generated by Django 4.2.18 on 2025-02-04 15:51

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion
import uuid


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ("authentication", "0004_platformuser_user_image_and_more"),
    ]

    operations = [
        migrations.AddField(
            model_name="platformorganization",
            name="abbreviation",
            field=models.CharField(blank=True, max_length=10, null=True),
        ),
        migrations.AddField(
            model_name="platformorganization",
            name="data",
            field=models.JSONField(default=dict),
        ),
        migrations.CreateModel(
            name="OrganizationCategory",
            fields=[
                ("created_at", models.DateTimeField(auto_now_add=True)),
                ("updated_at", models.DateTimeField(auto_now=True)),
                (
                    "id",
                    models.UUIDField(
                        default=uuid.uuid4,
                        editable=False,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                ("name", models.CharField(max_length=255, unique=True)),
                ("description", models.TextField(blank=True, null=True)),
                ("data", models.JSONField(default=dict)),
                (
                    "created_by",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="%(class)ss_created",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                (
                    "updated_by",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="%(class)ss_updated",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={
                "abstract": False,
            },
        ),
        migrations.AddField(
            model_name="platformorganization",
            name="org_category",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                to="authentication.organizationcategory",
            ),
        ),
    ]



import unittest
from collections import OrderedDict

from smoothglue.core.field_trie import ModifiedTrie, ModifiedTrieNode


class MockField:
    """
    A simplified mock for a serializer field.
    It can have nested fields accessible via get_fields() or be a simple value.
    """

    def __init__(self, is_nested=False, nested_fields=None):
        self._is_nested = is_nested
        if is_nested and nested_fields:
            self.fields = OrderedDict(nested_fields)
        elif is_nested:
            self.fields = OrderedDict()

    def get_fields(self):
        if self._is_nested:
            return self.fields
        raise AttributeError(
            "This mock field is not a nested serializer and has no 'get_fields' method."
        )


class MockBindingDict(OrderedDict):
    """
    A simplified mock for DRF's BindingDict.
    It behaves like an OrderedDict and holds MockField instances.
    It also needs a 'fields' attribute that points to itself for
    the pop operation in `compare_children_to_default_fields`.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.fields = self

    def get_fields(self):
        return self


class TestModifiedTrieNode(unittest.TestCase):
    def test_node_initialization_default(self):
        node = ModifiedTrieNode()
        self.assertIsNone(node.value)
        self.assertEqual(node.children, {})
        self.assertFalse(node.end)
        self.assertIsNone(node.fields_dict)

    def test_node_initialization_with_values(self):
        mock_fields = {"key": "value"}
        node = ModifiedTrieNode(
            value="test", end_of_field=True, fields_dict=mock_fields
        )
        self.assertEqual(node.value, "test")
        self.assertEqual(node.children, {})
        self.assertTrue(node.end)
        self.assertEqual(node.fields_dict, mock_fields)


class TestModifiedTrie(unittest.TestCase):
    def setUp(self):
        self.trie = ModifiedTrie()
        self.USER_NAME = "user.name"
        self.USER_USERNAME = "user.username"

    def test_initialization(self):
        self.assertIsNotNone(self.trie._ModifiedTrie__root)
        self.assertIsNone(self.trie._ModifiedTrie__root.value)
        self.assertEqual(self.trie._ModifiedTrie__root.children, {})

    def test_insert_single_field(self):
        self.trie.insert("id")
        self.assertIn("id", self.trie._ModifiedTrie__root.children)
        id_node = self.trie._ModifiedTrie__root.children["id"]
        self.assertEqual(id_node.value, "id")
        self.assertTrue(id_node.end)
        self.assertEqual(id_node.children, {})

    def test_insert_nested_field(self):
        self.trie.insert("user.profile.image")
        user_node = self.trie._ModifiedTrie__root.children.get("user")
        self.assertIsNotNone(user_node)
        self.assertEqual(user_node.value, "user")
        self.assertFalse(user_node.end)

        profile_node = user_node.children.get("profile")
        self.assertIsNotNone(profile_node)
        self.assertEqual(profile_node.value, "profile")
        self.assertFalse(profile_node.end)

        image_node = profile_node.children.get("image")
        self.assertIsNotNone(image_node)
        self.assertEqual(image_node.value, "image")
        self.assertTrue(image_node.end)
        self.assertEqual(image_node.children, {})

    def test_insert_multiple_fields_shared_prefix(self):
        self.trie.insert(self.USER_NAME)
        self.trie.insert("user.email")
        self.trie.insert("post.title")

        user_node = self.trie._ModifiedTrie__root.children.get("user")
        self.assertIsNotNone(user_node)
        self.assertIn("name", user_node.children)
        self.assertIn("email", user_node.children)
        self.assertTrue(user_node.children["name"].end)
        self.assertTrue(user_node.children["email"].end)

        post_node = self.trie._ModifiedTrie__root.children.get("post")
        self.assertIsNotNone(post_node)
        self.assertIn("title", post_node.children)
        self.assertTrue(post_node.children["title"].end)

    def test_insert_duplicate_field(self):
        self.trie.insert(self.USER_NAME)
        self.trie.insert(self.USER_NAME)  # Insert again

        user_node = self.trie._ModifiedTrie__root.children.get("user")
        name_node = user_node.children.get("name")
        self.assertTrue(name_node.end)
        # Ensure no duplicate nodes
        self.assertEqual(len(user_node.children), 1)

    def test_filter_requested_fields_empty_trie(self):
        # if nothing is requested, and nothing is default, nothing happens.
        # More accurately, if nothing requested, all default fields are popped).
        initial_fields = MockBindingDict({"id": MockField(), "name": MockField()})
        self.trie.filter_requested_fields(initial_fields)
        self.assertEqual(len(initial_fields), 0)

    def test_filter_requested_fields_no_matching_root_fields(self):
        self.trie.insert("nonexistent_field")
        initial_fields = MockBindingDict({"id": MockField(), "name": MockField()})
        self.trie.filter_requested_fields(initial_fields)
        # "id" and "name" were not requested through the trie (nonexistent_field was)
        # so "id" and "name" should be popped.
        self.assertEqual(len(initial_fields), 0)

    def test_filter_requested_fields_simple_match(self):
        self.trie.insert("name")
        initial_fields = MockBindingDict(
            {"id": MockField(), "name": MockField(), "email": MockField()}
        )
        self.trie.filter_requested_fields(initial_fields)
        self.assertIn("name", initial_fields)
        self.assertNotIn("id", initial_fields)
        self.assertNotIn("email", initial_fields)
        self.assertEqual(len(initial_fields), 1)

    def test_filter_requested_fields_nested_match(self):
        self.trie.insert("user.profile.image_url")
        self.trie.insert(self.USER_USERNAME)

        initial_fields = MockBindingDict(
            {
                "id": MockField(),
                "user": MockField(
                    is_nested=True,
                    nested_fields={
                        "username": MockField(),
                        "email": MockField(),  # Not requested
                        "profile": MockField(
                            is_nested=True,
                            nested_fields={
                                "image_url": MockField(),
                                "bio": MockField(),  # Not requested
                            },
                        ),
                    },
                ),
                "post_count": MockField(),  # Not requested
            }
        )

        self.trie.filter_requested_fields(initial_fields)

        self.assertNotIn("id", initial_fields)
        self.assertNotIn("post_count", initial_fields)
        self.assertIn("user", initial_fields)

        user_field_data = initial_fields["user"]
        self.assertIn("username", user_field_data.fields)
        self.assertNotIn("email", user_field_data.fields)
        self.assertIn("profile", user_field_data.fields)

        profile_field_data = user_field_data.fields["profile"]
        self.assertIn("image_url", profile_field_data.fields)
        self.assertNotIn("bio", profile_field_data.fields)

    def test_filter_requested_fields_partial_nested_match(self):

        self.trie.insert("user")

        initial_fields = MockBindingDict(
            {
                "user": MockField(
                    is_nested=True,
                    nested_fields={"username": MockField(), "email": MockField()},
                )
            }
        )

        self.trie.filter_requested_fields(initial_fields)

        self.assertIn("user", initial_fields)
        user_field_data = initial_fields["user"]
        self.assertIn("username", user_field_data.fields)
        self.assertIn("email", user_field_data.fields)

    def test_filter_requested_fields_request_parent_and_specific_child(self):
        self.trie.insert(self.USER_USERNAME)

        initial_fields = MockBindingDict(
            {
                "user": MockField(
                    is_nested=True,
                    nested_fields={
                        "username": MockField(),
                        "email": MockField(),  # This should be pruned
                        "status": MockField(),  # This should be pruned
                    },
                ),
                "other_field": MockField(),  # This should be pruned
            }
        )

        self.trie.filter_requested_fields(initial_fields)

        self.assertNotIn("other_field", initial_fields)
        self.assertIn("user", initial_fields)
        user_field_data = initial_fields["user"]

        self.assertIn("username", user_field_data.fields)
        self.assertNotIn("email", user_field_data.fields)
        self.assertNotIn("status", user_field_data.fields)

    def test_filter_requested_fields_non_existent_nested_field_in_trie(self):
        self.trie.insert("user.profile.non_existent_detail")

        initial_fields = MockBindingDict(
            {
                "user": MockField(
                    is_nested=True,
                    nested_fields={
                        "username": MockField(),
                        "profile": MockField(
                            is_nested=True, nested_fields={"image_url": MockField()}
                        ),
                    },
                )
            }
        )
        self.trie.filter_requested_fields(initial_fields)

        self.assertIn("user", initial_fields)
        user_field_data = initial_fields["user"]
        self.assertIn("profile", user_field_data.fields)
        profile_field_data = user_field_data.fields["profile"]

        # image_url was not requested, so it's removed.
        self.assertNotIn("image_url", profile_field_data.fields)
        self.assertEqual(len(profile_field_data.fields), 0)

    def test_filter_requested_fields_field_without_get_fields(self):
        self.trie.insert("problem_field.child")

        class MockNonNestedFieldWithChildrenInTrie:
            # This mock won't have 'get_fields'.
            # It also won't have a 'fields' attribute like MockField(is_nested=True)
            # For the pop operation, the original code relies on current.fields_dict.fields.pop()
            # So, the parent (problem_field's dict) must have .fields
            pass

        initial_fields = MockBindingDict(
            # problem_field does not have .get_fields()
            {
                "problem_field": MockNonNestedFieldWithChildrenInTrie(),
                "another_field": MockField(),
            }
        )
        # The 'problem_field' node in the trie will have a child 'child'.
        # When processing 'problem_field',
        # current.fields_dict will be MockNonNestedFieldWithChildrenInTrie.
        # hasattr(current.fields_dict, "get_fields") will be False.
        # The `continue` should be hit. 'another_field' should be popped.
        self.trie.filter_requested_fields(initial_fields)

        self.assertIn(
            "problem_field", initial_fields
        )  # Kept because it was in trie, but children not processed
        self.assertNotIn("another_field", initial_fields)  # Removed as not requested

    def test_compare_children_to_default_fields_all_children_match(self):
        # Setup parent node and its children in the trie
        parent_node = ModifiedTrieNode("parent")
        child1_node = ModifiedTrieNode("child1")
        child2_node = ModifiedTrieNode("child2")
        parent_node.children = {"child1": child1_node, "child2": child2_node}

        parent_fields_dict_fields_attr = MockBindingDict(
            {
                "child1": MockField(
                    is_nested=True, nested_fields={"sub1": MockField()}
                ),
                "child2": MockField(),
                "extra_in_dict": MockField(),  # This should be removed
            }
        )

        parent_node.fields_dict = MockField(
            is_nested=True, nested_fields=parent_fields_dict_fields_attr
        )
        parent_node.fields_dict.fields = parent_fields_dict_fields_attr

        default_db_fields = {"child1", "child2", "extra_in_dict"}
        field_queue = []

        self.trie.compare_children_to_default_fields(
            parent_node, default_db_fields, field_queue
        )

        self.assertIsNotNone(child1_node.fields_dict)
        self.assertEqual(
            child1_node.fields_dict, parent_fields_dict_fields_attr["child1"]
        )
        self.assertIsNotNone(child2_node.fields_dict)
        self.assertEqual(
            child2_node.fields_dict, parent_fields_dict_fields_attr["child2"]
        )

        self.assertIn(child1_node, field_queue)
        self.assertIn(child2_node, field_queue)

        self.assertNotIn("extra_in_dict", parent_node.fields_dict.fields)
        self.assertIn("child1", parent_node.fields_dict.fields)
        self.assertIn("child2", parent_node.fields_dict.fields)

    def test_compare_children_to_default_fields_some_children_not_in_defaults(self):
        parent_node = ModifiedTrieNode("parent")
        child1_node = ModifiedTrieNode("child1")  # in defaults
        child_not_in_defaults_node = ModifiedTrieNode(
            "child_not_in_defaults"
        )  # NOT in defaults
        parent_node.children = {
            "child1": child1_node,
            "child_not_in_defaults": child_not_in_defaults_node,
        }

        parent_fields_dict_fields_attr = MockBindingDict(
            {
                "child1": MockField(),
                "other_field_in_dict": MockField(),
            }
        )
        parent_node.fields_dict = MockField(
            is_nested=True, nested_fields=parent_fields_dict_fields_attr
        )
        parent_node.fields_dict.fields = parent_fields_dict_fields_attr

        default_db_fields = {
            "child1",
            "other_field_in_dict",
        }  # child_not_in_defaults is not here
        field_queue = []

        self.trie.compare_children_to_default_fields(
            parent_node, default_db_fields, field_queue
        )

        self.assertIsNotNone(child1_node.fields_dict)
        self.assertIsNone(
            child_not_in_defaults_node.fields_dict
        )  # Not in defaults, so not processed

        self.assertIn(child1_node, field_queue)
        self.assertNotIn(child_not_in_defaults_node, field_queue)

        # 'other_field_in_dict' was in default_db_fields
        # but not in parent_node.children, so it's removed.
        self.assertNotIn("other_field_in_dict", parent_node.fields_dict.fields)
        self.assertIn("child1", parent_node.fields_dict.fields)

    def test_filter_root_level_fields_correctly_when_is_root_true(self):
        self.trie.insert("name")
        # initial_fields is the root fields_dict
        initial_fields = MockBindingDict(
            {
                "id": MockField(),
                "name": MockField(is_nested=False),  # Mark as not nested for clarity
                "email": MockField(),
            }
        )

        # Ensure the root node's fields_dict is set to initial_fields
        self.trie._ModifiedTrie__root.fields_dict = initial_fields

        self.trie.filter_requested_fields(initial_fields)

        self.assertIn("name", initial_fields)
        self.assertNotIn(
            "id", initial_fields
        )  # Popped by compare_children called on root
        self.assertNotIn(
            "email", initial_fields
        )  # Popped by compare_children called on root
        self.assertEqual(len(initial_fields), 1)

        name_node = self.trie._ModifiedTrie__root.children["name"]
        self.assertEqual(name_node.fields_dict, initial_fields["name"])

    def test_filter_nested_fields_with_get_fields(self):
        self.trie.insert(self.USER_USERNAME)
        nested_user_fields = OrderedDict(
            {"username": MockField(), "email": MockField()}
        )
        user_mock_field = MockField(is_nested=True, nested_fields=nested_user_fields)

        initial_fields = MockBindingDict({"user": user_mock_field, "id": MockField()})

        self.trie.filter_requested_fields(initial_fields)

        self.assertIn("user", initial_fields)
        self.assertNotIn("id", initial_fields)

        user_data_after_filter = initial_fields["user"]
        self.assertIn("username", user_data_after_filter.fields)
        self.assertNotIn("email", user_data_after_filter.fields)

        user_node = self.trie._ModifiedTrie__root.children["user"]
        username_node = user_node.children["username"]
        self.assertEqual(username_node.fields_dict, nested_user_fields["username"])



"""Tests for views/_views.py to improve coverage."""

from unittest.mock import MagicMock, patch

import pytest
from django.conf import settings
from django.test import RequestFactory
from rest_framework import status
from rest_framework.response import Response

from smoothglue.core.views._views import (
    CustomDataSchema,
    ListModelViewSet,
    get_app_info,
)


class TestListModelViewSetExceptionHandler:
    """Test exception handling in ListModelViewSet."""

    def test_exception_handler_with_default_response(self):
        """Test exception handler when default handler returns response."""
        exc = ValueError("Test error")
        context = {"request": MagicMock()}
        context["request"].method = "POST"

        # Mock default handler to return a response
        with patch(
            "smoothglue.core.views._views.default_exception_handler"
        ) as mock_handler:
            mock_response = MagicMock()
            mock_handler.return_value = mock_response

            result = ListModelViewSet.exception_handler(exc, context)

            assert result == mock_response
            mock_handler.assert_called_once_with(exc, context)

    def test_exception_handler_with_no_default_response(self):
        """Test exception handler when default handler returns None."""
        exc = ValueError("Test error")
        context = {"request": MagicMock()}
        context["request"].method = "DELETE"

        # Mock default handler to return None
        with patch(
            "smoothglue.core.views._views.default_exception_handler"
        ) as mock_handler:
            mock_handler.return_value = None

            with patch("smoothglue.core.views._views.logger") as mock_logger:
                result = ListModelViewSet.exception_handler(exc, context)

                # Check response
                assert isinstance(result, Response)
                assert result.status_code == status.HTTP_400_BAD_REQUEST
                assert result.data == {"error": "Test error"}

                # Check logging
                mock_logger.error.assert_called_once_with(
                    "Error %s object - %s: %s",
                    "deleting",
                    "ValueError",
                    "Test error",
                    exc_info=True,
                )

    def test_exception_handler_unknown_method(self):
        """Test exception handler with unknown HTTP method."""
        exc = ValueError("Test error")
        context = {"request": MagicMock()}
        context["request"].method = "HEAD"  # Not in error_verbs

        with patch(
            "smoothglue.core.views._views.default_exception_handler", return_value=None
        ):
            with patch("smoothglue.core.views._views.logger") as mock_logger:
                result = ListModelViewSet.exception_handler(exc, context)

                # Check that empty string is used for unknown method
                mock_logger.error.assert_called_once_with(
                    "Error %s object - %s: %s",
                    "",  # Empty string for unknown method
                    "ValueError",
                    "Test error",
                    exc_info=True,
                )

    def test_get_exception_handler(self):
        """Test get_exception_handler method."""
        viewset = ListModelViewSet()
        handler = viewset.get_exception_handler()
        assert handler == viewset.exception_handler


class TestCustomDataSchema:
    """Test CustomDataSchema class."""

    def test_init_with_data_schema(self):
        """Test CustomDataSchema initialization with data_schema."""
        data_schema = {"type": "object", "properties": {}}

        schema = CustomDataSchema(data_schema=data_schema)

        assert schema.data_schema == data_schema

    def test_map_field_data_field(self):
        """Test map_field method for 'data' field."""
        data_schema = {"type": "object", "properties": {}}
        schema = CustomDataSchema(data_schema=data_schema)

        # Mock field with field_name = "data"
        field = MagicMock()
        field.field_name = "data"

        result = schema.map_field(field)

        assert result == data_schema

    def test_map_field_other_field(self):
        """Test map_field method for non-'data' field."""
        data_schema = {"type": "object", "properties": {}}
        schema = CustomDataSchema(data_schema=data_schema)

        # Mock field with different field_name
        field = MagicMock()
        field.field_name = "other_field"

        # Mock parent map_field method
        with patch.object(CustomDataSchema.__bases__[0], "map_field") as mock_super:
            mock_super.return_value = {"type": "string"}

            result = schema.map_field(field)

            assert result == {"type": "string"}
            mock_super.assert_called_once_with(field)


# Note: get_app_info tests are omitted since CLASSIFICATION_LEVEL and CLASSIFICATION_MESSAGE
# are project-level settings that should be provided by the Django project using this package,
# not by the package itself. Testing these would require mocking settings which is not
# appropriate for a reusable Django app.



from unittest.mock import MagicMock, patch

import pytest
from rest_framework.views import Request

from smoothglue.core.views import AsyncListModelViewSet, ListModelViewSet


class ChildListModelViewSet(ListModelViewSet):
    function_called: bool = False

    def get_queryset(self):
        return None

    def get_serializer_class(self):
        return MagicMock()

    def list(self, request, *args, **kwargs):
        self.function_called = True
        return super().list(request, *args, **kwargs)

    def create(self, request, *args, **kwargs):
        self.function_called = True
        return super().create(request, *args, **kwargs)

    def update(self, request, *args, **kwargs):
        self.function_called = True
        return super().update(request, *args, **kwargs)

    def partial_update(self, request, *args, **kwargs):
        self.function_called = True
        return super().partial_update(request, *args, **kwargs)

    def destroy(self, request, *args, **kwargs):
        self.function_called = True
        return super().destroy(request, *args, **kwargs)


class SubChildListModelViewSet(ChildListModelViewSet):
    def get_queryset(self):
        return None

    def get_serializer_class(self):
        return MagicMock()

    def list(self, request, *args, **kwargs):
        return super().list(request, *args, **kwargs)

    def create(self, request, *args, **kwargs):
        return super().create(request, *args, **kwargs)

    def update(self, request, *args, **kwargs):
        return super().update(request, *args, **kwargs)

    def partial_update(self, request, *args, **kwargs):
        return super().partial_update(request, *args, **kwargs)

    def destroy(self, request, *args, **kwargs):
        return super().destroy(request, *args, **kwargs)


@pytest.mark.parametrize(
    "calling_function_name, patch_target",
    [
        (
            "list",
            "rest_framework.viewsets.ModelViewSet.list",
        ),
        (
            "create",
            "rest_framework.viewsets.ModelViewSet.perform_create",
        ),
        (
            "update",
            "rest_framework.viewsets.ModelViewSet.update",
        ),
        (
            "partial_update",
            "rest_framework.viewsets.ModelViewSet.partial_update",
        ),
        (
            "destroy",
            "rest_framework.viewsets.ModelViewSet.destroy",
        ),
    ],
)
@pytest.mark.django_db
def test_child_function_call(calling_function_name, patch_target):
    """
    Given a class that inherits multiple times from ListModelViewSet
    When we can any of the core `view` methods
    Then overridden functions between the child and parent should still be called
    """

    sub_child_view = SubChildListModelViewSet()
    sub_child_view.request = None
    sub_child_view.format_kwarg = None
    assert sub_child_view.function_called is False

    mock_request = MagicMock()

    with patch(patch_target) as mock_list_func:
        getattr(sub_child_view, calling_function_name)(mock_request)
        mock_list_func.assert_called()
        assert sub_child_view.function_called


@pytest.mark.parametrize(
    "calling_function_name, patch_target",
    [
        (
            "alist",
            "adrf.viewsets.ModelViewSet.alist",
        ),
        (
            "acreate",
            "adrf.viewsets.ModelViewSet.perform_create",
        ),
        (
            "aupdate",
            "adrf.viewsets.ModelViewSet.aupdate",
        ),
        (
            "partial_aupdate",
            "adrf.viewsets.ModelViewSet.partial_aupdate",
        ),
        (
            "adestroy",
            "adrf.viewsets.ModelViewSet.destroy",
        ),
    ],
)
@pytest.mark.asyncio
@pytest.mark.django_db
async def test_async_views_functions_called(
    calling_function_name, patch_target, get_test_model, get_test_async_serializer
):
    """
    Given a class that inherits multiple times from AsyncListModelViewSet
    When we can any of the core `view` methods
    Then overridden functions between the child and parent should still be called
    """
    SomeModel = get_test_model
    SomeSerializer = get_test_async_serializer

    class AsyncChildListModelViewSet(AsyncListModelViewSet):
        function_called: bool = False

        async def get_queryset(self):
            return None

        async def alist(self, request, *args, **kwargs):
            self.function_called = True
            return await super().alist(request, *args, **kwargs)

        async def acreate(self, request, *args, **kwargs):
            self.function_called = True
            return await super().acreate(request, *args, **kwargs)

        async def aupdate(self, request, *args, **kwargs):
            self.function_called = True
            return await super().aupdate(request, *args, **kwargs)

        async def partial_aupdate(self, request, *args, **kwargs):
            self.function_called = True
            return await super().partial_aupdate(request, *args, **kwargs)

        async def adestroy(self, request, *args, **kwargs):
            self.function_called = True
            return await super().adestroy(request, *args, **kwargs)

    class AsyncSubChildListModelViewSet(AsyncChildListModelViewSet):
        serializer_class = SomeSerializer

        def get_queryset(self):
            return None

        async def alist(self, request, *args, **kwargs):
            return await super().alist(request, *args, **kwargs)

        async def acreate(self, request, *args, **kwargs):
            return await super().acreate(request, *args, **kwargs)

        async def aupdate(self, request, *args, **kwargs):
            return await super().aupdate(request, *args, **kwargs)

        async def partial_aupdate(self, request, *args, **kwargs):
            return await super().partial_aupdate(request, *args, **kwargs)

        async def adestroy(self, request, *args, **kwargs):
            return await super().adestroy(request, *args, **kwargs)

        def get_exception_handler(self):
            return super().get_exception_handler()

    async_sub_child_view = AsyncSubChildListModelViewSet()
    async_sub_child_view.request = None
    async_sub_child_view.format_kwarg = None
    async_sub_child_view.kwargs = {"pk": 1}
    assert async_sub_child_view.function_called is False

    mock_request = MagicMock(spec=Request)
    mock_request.data = {"id": 1, "name": "some name"}

    with patch(
        "rest_framework.generics.get_object_or_404",
        return_value=SomeModel(pk=1, name="some name"),
    ):
        with patch(patch_target) as mock_func:
            await getattr(async_sub_child_view, calling_function_name)(mock_request)
    mock_func.assert_called()
    assert async_sub_child_view.function_called is True



from typing import Any, Dict, List

from django.conf import settings


def check_kwargs(list_keywords: List[str], kwargs: Dict[str, Any]) -> None:
    """
    Check if the provided keywords exist in the kwargs dictionary.

    This function iterates over each keyword in the list_keywords and asserts
    if each keyword exists as a key in the kwargs dictionary. An assertion error
    is raised if a keyword is not found as a key in the kwargs.

    Parameters:
    - list_keywords : List[str]
        A list of keywords to be checked in the kwargs.

    - kwargs : Dict[str, Any]
        A dictionary of keyword arguments in which the keywords
        from list_keywords will be checked.

    Returns:
    - None

    Raises:
    - AssertionError:
        If a keyword from list_keywords is not found as a key in kwargs.
    """

    for keyword in list_keywords:
        assert (
            keyword in kwargs
        ), f"{keyword} does not exist in kwargs keys:\n{list(kwargs.keys())}"


def get_setting(name, app_defaults):
    """
    Retrieves a setting value from the Django project's settings, falling back to the app's
    default settings if the setting is not found in the project's settings.

    Parameters:
    name (str): The name of the setting.
    app_defaults (module): The defaults module of the app where default settings are defined.

    Returns:
    The value of the setting from the project's settings or the app's default settings.
    """
    return getattr(settings, name, getattr(app_defaults, name))



from django.contrib import admin

from smoothglue.tracker.models import APIChangeLog, AppErrorLog


@admin.register(APIChangeLog)
class APIChangeLogAdmin(admin.ModelAdmin):
    search_fields = ["full_path"]
    list_display = ("full_path", "username", "method", "timestamp", "data")
    list_filter = (
        "username",
        "method",
        "timestamp",
    )
    readonly_fields = ["username", "full_path", "method", "timestamp", "data", "params"]
    ordering = ("-timestamp",)

    def has_add_permission(self, request):
        return False

    def has_change_permission(self, request, obj=None):
        return False

    def has_view_permission(self, request, obj=None):
        return request.user.is_superuser


@admin.register(AppErrorLog)
class AppErrorLogAdmin(admin.ModelAdmin):
    search_fields = ["user"]
    list_display = ("level", "message", "user", "timestamp")
    list_filter = (
        "user",
        "level",
        "timestamp",
    )
    readonly_fields = [
        "user",
        "level",
        "message",
        "stack_trace",
        "client_data",
        "timestamp",
    ]
    ordering = ("-timestamp",)

    def has_add_permission(self, request):
        return False

    def has_change_permission(self, request, obj=None):
        return False

    def has_view_permission(self, request, obj=None):
        return request.user.is_superuser



from django.conf import settings
from django.contrib.auth.middleware import PersistentRemoteUserMiddleware
from django.utils.cache import patch_vary_headers
from django.utils.deprecation import MiddlewareMixin


# pylint: disable=R0205
class DoNotTrackMiddleware(MiddlewareMixin):
    @classmethod
    def process_request(cls, request):
        """
        Sets flag request.DNT based on DNT HTTP header.
        """
        if "HTTP_DNT" in request.META and request.META["HTTP_DNT"] == "1":
            request.DNT = True
        else:
            request.DNT = False

    @classmethod
    def process_response(cls, request, response):
        """
        Adds a "Vary" header for DNT, useful for caching.
        """
        patch_vary_headers(response, ["DNT"])

        return response


class CACMiddleware(PersistentRemoteUserMiddleware):
    """
    Middleware that takes in the Authorization header to be used for
    authentication
    """

    def __init__(self, get_response):
        super().__init__(get_response)
        self.header = settings.REMOTE_USER_HEADER



"""
This is required to be a separate module for pylint.
Since pylint isn't able to utilize the boot_django utility a "real" django settings module is
required.
"""

import os

BASE_DIR = os.path.abspath(os.path.join(os.path.dirname(__file__), "smoothglue"))
DEBUG = True
DATABASES = {
    "default": {
        "ENGINE": "django.db.backends.sqlite3",
        "NAME": os.path.join(BASE_DIR, "db.sqlite3"),
    }
}
INSTALLED_APPS = [
    "django.contrib.contenttypes",
    "django.contrib.sessions",
    "django.contrib.auth",
    "django.contrib.admin",
    "smoothglue.core",
    "smoothglue.tracker",
    "smoothglue.authentication",
]
TIME_ZONE = "UTC"
USE_TZ = True
ROOT_URLCONF = "smoothglue.authentication.urls"
AUTH_USER_MODEL = "authentication.PlatformUser"
SECRET_KEY = "TEST"
REMOTE_USER_HEADER = "HTTP_X_USERNAME"



from django.contrib import admin
from django.db import models
from django_json_widget.widgets import JSONEditorWidget


class DataAdmin(admin.ModelAdmin):
    """
    This provides an easier WYSIWYG JSON Editor for JSONField
    """

    formfield_overrides = {
        models.JSONField: {"widget": JSONEditorWidget},
    }


class BaseAuditModelAdmin(DataAdmin):
    """
    This base admin class will disable the audit fields and will auto populate
    the audit fields based on the request users.
    """

    readonly_fields = ("created_by", "updated_by", "created_at", "updated_at")

    def save_model(self, request, obj, form, change):
        if not change:
            obj.created_by = request.user
        obj.updated_by = request.user
        return super().save_model(request, obj, form, change)

    def save_formset(self, request, form, formset, change):
        instances = formset.save(commit=False)
        for instance in instances:
            if not hasattr(instance, "created_by"):
                instance.created_by = request.user
            instance.updated_by = request.user
            instance.save()



# Generated by Django 4.2.18 on 2025-01-17 05:49

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ("authentication", "0003_platformuser_organizations_and_more"),
    ]

    operations = [
        migrations.AddField(
            model_name="platformuser",
            name="user_image",
            field=models.ImageField(blank=True, null=True, upload_to=None),
        ),
        migrations.AlterField(
            model_name="organizationmember",
            name="platform_user",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                to="authentication.platformuser",
            ),
        ),
        migrations.AlterField(
            model_name="platformorganization",
            name="platform_users",
            field=models.ManyToManyField(
                default=None,
                through="authentication.OrganizationMember",
                to="authentication.platformuser",
            ),
        ),
    ]



from unittest.mock import Mock, patch

from django.conf import settings
from django.test import TestCase, override_settings
from django.contrib.auth.models import Group
from django.contrib.auth import get_user_model

from smoothglue.authentication.models import (
    PlatformOrganization,
    is_group_security_enabled,
    post_save_user_signal_handler,
    post_save_org_signal_handler,
    post_save_org_member_signal_handler,
    post_delete_org_member_signal_handler,
)

User = get_user_model()


class UserSignalTests(TestCase):
    """Tests for signals related to the PlatformUser model."""

    def test_new_user_added_to_default_group(self):
        """
        Verify a new user is added to the default 'Participant' group if it exists.
        """
        participant_group, _ = Group.objects.get_or_create(name="Participant")
        user = User.objects.create_user(username="testuser1", email="test1@example.com")
        self.assertIn(participant_group, user.groups.all())

    @override_settings(GLOBAL_USER_GROUP_NAME="CustomGlobalGroup")
    def test_new_user_added_to_settings_group(self):
        """
        Verify a new user is added to the group specified in settings.
        """
        custom_group, _ = Group.objects.get_or_create(name="CustomGlobalGroup")
        user = User.objects.create_user(username="testuser2", email="test2@example.com")
        self.assertIn(custom_group, user.groups.all())

    def test_user_not_added_if_group_does_not_exist(self):
        """
        Verify the signal does nothing if the target group isn't in the database.
        """
        Group.objects.filter(name="Participant").delete()
        user = User.objects.create_user(username="testuser3", email="test3@example.com")
        self.assertEqual(user.groups.count(), 0)

    def test_signal_does_not_fire_on_user_update(self):
        """
        Verify the signal only runs on user creation, not on updates.
        """
        user = User.objects.create_user(username="testuser4", email="test4@example.com")
        self.assertEqual(user.groups.count(), 0)
        user.first_name = "Updated"
        user.save()
        self.assertEqual(user.groups.count(), 0)


@override_settings(ENABLE_GROUP_SECURITY=True)
class OrganizationSignalTests(TestCase):
    """
    Tests for signals related to PlatformOrganization and its members.
    """

    def setUp(self):
        self.user = User.objects.create_user(
            username="orguser", email="orguser@example.com"
        )
        self.org = PlatformOrganization.objects.create(name="Test Corp")
        self.org.refresh_from_db()

    def test_org_creation_creates_linked_group(self):
        self.assertIsNotNone(self.org.group)
        self.assertEqual(self.org.group.name, self.org.name)
        self.assertTrue(Group.objects.filter(name="Test Corp").exists())

    @override_settings(ENABLE_GROUP_SECURITY=False)
    def test_org_creation_does_not_create_group_if_disabled(self):
        org_no_group = PlatformOrganization.objects.create(name="No Group Corp")
        self.assertIsNone(org_no_group.group)

    def test_adding_user_to_org_adds_to_group(self):
        self.assertNotIn(self.org.group, self.user.groups.all())
        self.org.platform_users.add(self.user)
        self.assertIn(self.org.group, self.user.groups.all())

    def test_removing_user_from_org_removes_from_group(self):
        self.org.platform_users.add(self.user)
        self.assertIn(self.org.group, self.user.groups.all())
        self.org.platform_users.remove(self.user)
        self.assertNotIn(self.org.group, self.user.groups.all())


class SignalHandlerUnitTests(TestCase):
    """
    Unit tests that call signal handler functions directly to test their
    internal logic for full coverage.
    """

    def test_is_group_security_enabled_helper(self):
        """Tests the `is_group_security_enabled` helper function."""
        with override_settings(ENABLE_GROUP_SECURITY=True):
            self.assertTrue(is_group_security_enabled())
        with override_settings(ENABLE_GROUP_SECURITY=False):
            self.assertFalse(is_group_security_enabled())
        # Test default behavior when setting is not present
        with override_settings():
            delattr(settings, "ENABLE_GROUP_SECURITY")
            self.assertFalse(is_group_security_enabled())

    @patch("smoothglue.authentication.models.Group.objects.filter")
    def test_post_save_user_handler_logic(self, mock_group_filter):
        """Unit tests the user creation signal handler."""
        mock_user = Mock(spec=User)
        mock_group = Mock(spec=Group)
        # Simulate the group existing
        mock_group_filter.return_value.first.return_value = mock_group

        post_save_user_signal_handler(instance=mock_user, created=True)
        mock_user.groups.add.assert_called_once_with(mock_group)
        mock_user.save.assert_called_once()

        mock_user.reset_mock()
        post_save_user_signal_handler(instance=mock_user, created=False)
        mock_user.groups.add.assert_not_called()

    @patch("smoothglue.authentication.models.Group.objects.create")
    @override_settings(ENABLE_GROUP_SECURITY=True)
    def test_post_save_org_handler_logic(self, mock_group_create):
        """Unit tests the organization creation signal handler."""
        mock_org = Mock(spec=PlatformOrganization, name="New Org")

        post_save_org_signal_handler(instance=mock_org, created=True)

        self.assertTrue(mock_org.save.called)

    def test_organization_member_signal_handlers(self):
        """Unit tests the handlers for OrganizationMember creation and deletion."""
        mock_user = Mock(spec=User)
        mock_group = Mock(spec=Group)
        mock_org = Mock(spec=PlatformOrganization, group=mock_group)
        mock_member = Mock(platform_user=mock_user, platform_organization=mock_org)

        with override_settings(ENABLE_GROUP_SECURITY=True):
            # Test add signal
            post_save_org_member_signal_handler(instance=mock_member, created=True)
            mock_user.groups.add.assert_called_once_with(mock_group)

            # Test delete signal
            post_delete_org_member_signal_handler(instance=mock_member)
            mock_user.groups.remove.assert_called_once_with(mock_group)



from django.test import RequestFactory, TestCase, override_settings
from django.http import HttpResponse
from django.urls import path
from smoothglue.authentication.models import PlatformUser
from smoothglue.core.middleware import (
    DoNotTrackMiddleware,
)


def dummy_view(request):
    return HttpResponse(status=200)


urlpatterns = [
    path("test/", dummy_view),
]


@override_settings(
    INSTALLED_APPS=[
        "django.contrib.auth",
        "django.contrib.contenttypes",
        "django.contrib.sessions",
        "smoothglue.core",
        "smoothglue.authentication",
    ],
    REMOTE_USER_HEADER="HTTP_X_USERNAME",
    AUTHENTICATION_BACKENDS=["django.contrib.auth.backends.RemoteUserBackend"],
    MIDDLEWARE=[
        "django.contrib.sessions.middleware.SessionMiddleware",
        "django.contrib.auth.middleware.AuthenticationMiddleware",
        "smoothglue.core.middleware.CACMiddleware",
    ],
    ROOT_URLCONF="smoothglue.core.tests.test_middleware",
)
class CACMiddlewareTest(TestCase):
    def setUp(self):
        self.user = PlatformUser.objects.create(username="testuser")

    def test_user_is_authenticated_with_valid_header(self):
        response = self.client.get("/test/", HTTP_X_USERNAME="testuser")
        self.assertEqual(response.wsgi_request.user, self.user)

    def test_user_is_anonymous_with_no_header(self):
        response = self.client.get("/test/")
        self.assertTrue(response.wsgi_request.user.is_anonymous)


class DoNotTrackMiddlewareTest(TestCase):
    """Tests for the DoNotTrackMiddleware."""

    def setUp(self):
        """Set up the request factory."""
        self.factory = RequestFactory()

    def test_dnt_header_enabled(self):
        """Tests that request.DNT is True when the DNT header is '1'."""
        request = self.factory.get("/", HTTP_DNT="1")
        DoNotTrackMiddleware.process_request(request)
        self.assertTrue(request.DNT)

    def test_dnt_header_disabled(self):
        """Tests that request.DNT is False when the DNT header is '0'."""
        request = self.factory.get("/", HTTP_DNT="0")
        DoNotTrackMiddleware.process_request(request)
        self.assertFalse(request.DNT)

    def test_dnt_header_absent(self):
        """Tests that request.DNT is False when the DNT header is not present."""
        request = self.factory.get("/")
        DoNotTrackMiddleware.process_request(request)
        self.assertFalse(request.DNT)

    def test_vary_header_is_added_to_response(self):
        """Tests that 'DNT' is added to the 'Vary' header of the response."""
        request = self.factory.get("/")
        response = HttpResponse()

        processed_response = DoNotTrackMiddleware.process_response(request, response)

        self.assertEqual(processed_response.get("Vary"), "DNT")



import logging

from adrf import viewsets
from adrf.mixins import get_data
from django.db.models.base import sync_to_async
from django.db.models.query import transaction
from drf_spectacular.utils import OpenApiParameter, extend_schema
from rest_framework import status
from rest_framework.response import Response
from rest_framework.views import exception_handler as default_exception_handler

from smoothglue.tracker.utils import create_change_log

logger = logging.getLogger(__name__)


class AsyncListModelViewSet(
    viewsets.ModelViewSet
):  # pylint: disable=too-many-ancestors
    """
    Thin wrapper around the smoothglue.views.ListModelViewSet
    to provide async functionality through the standard View functions.
    """

    @extend_schema(
        methods=["get"],
        parameters=[
            OpenApiParameter(
                name="fields",
                type={"type": "array", "items": {"type": "string"}},
                location=OpenApiParameter.QUERY,
                description="Fields to include in the response",
            ),
        ],
    )
    async def alist(self, *args, **kwargs):
        return await super().alist(*args, **kwargs)

    def get_exception_handler(
        self,
    ):
        return self.exception_handler

    @staticmethod
    def exception_handler(exc, context):
        response = default_exception_handler(exc, context)
        if response is not None:
            return response

        error_verbs = {
            "DELETE": "deleting",
            "PUT": "modifying",
            "PATCH": "modifying",
            "POST": "creating",
        }
        logger.error(
            "Error %s object - %s: %s",
            error_verbs.get(context["request"].method, ""),
            type(exc).__name__,
            str(exc),
            exc_info=True,
        )
        return Response({"error": str(exc)}, status=status.HTTP_400_BAD_REQUEST)

    @transaction.atomic
    def perform_create_atomic(self, serializer):
        self.perform_create(serializer)

    async def acreate(self, request, *args, **kwargs):
        """
        Creates a single object or a list of objects. All operations are wrapped in an atomic
        transaction to ensure data integrity. A change log is created after the operation.
        """
        is_many = isinstance(request.data, list)
        serializer = self.get_serializer(data=request.data, many=is_many)
        await sync_to_async(serializer.is_valid)(raise_exception=True)
        await sync_to_async(self.perform_create_atomic)(serializer)
        create_change_log(request)
        data = await get_data(serializer)
        headers = self.get_success_headers(data)
        return Response(data, status=status.HTTP_201_CREATED, headers=headers)

    async def aupdate(self, request, *args, **kwargs):
        response = await super().aupdate(request, *args, **kwargs)
        create_change_log(request)
        return response

    async def partial_aupdate(self, request, *args, **kwargs):
        response = await super().partial_aupdate(request, *args, **kwargs)
        create_change_log(request)
        return response

    @transaction.atomic
    def destroy_atomic(self, request, *args, **kwargs):
        return super().destroy(request, *args, **kwargs)

    async def adestroy(self, request, *args, **kwargs):
        response = await sync_to_async(self.destroy_atomic)(request, *args, **kwargs)
        create_change_log(request)
        return response



from django.contrib.admin.sites import AdminSite
from django.contrib.auth import get_user_model
from django.db import models
from django.test import RequestFactory, TestCase

from smoothglue.authentication.admin import (
    OrganizationCategoryAdmin,
    OrganizationMembersInline,
    PlatformOrganizationAdmin,
    PlatformOrganizationMembersInline,
    PlatformOrganizationsInline,
    UserAdmin,
)
from smoothglue.authentication.models import OrganizationCategory, PlatformOrganization

User = get_user_model()


class MockRequest:
    pass


class TestUserAdmin(TestCase):
    def setUp(self):
        self.site = AdminSite()
        self.user_admin = UserAdmin(User, self.site)
        self.request = RequestFactory().get("/admin/authentication/user/")

    def test_readonly_fields(self):
        self.assertEqual(self.user_admin.readonly_fields, ("username", "email"))

    def test_list_display(self):
        self.assertEqual(
            self.user_admin.list_display,
            ("username", "email", "first_name", "last_name", "is_staff"),
        )

    def test_search_fields(self):
        self.assertEqual(
            self.user_admin.search_fields,
            ("username", "email", "first_name", "last_name"),
        )

    def test_list_filter(self):
        self.assertEqual(
            self.user_admin.list_filter,
            ("is_staff", "is_superuser", "is_active", "groups"),
        )

    def test_formfield_overrides(self):
        self.assertIn(models.JSONField, self.user_admin.formfield_overrides)

    def test_fieldsets(self):
        self.assertEqual(len(self.user_admin.fieldsets), 5)

    def test_inlines(self):
        self.assertIn(OrganizationMembersInline, self.user_admin.inlines)


class TestPlatformOrganizationAdmin(TestCase):
    def setUp(self):
        self.site = AdminSite()
        self.platform_org_admin = PlatformOrganizationAdmin(
            PlatformOrganization, self.site
        )
        self.request = RequestFactory().get(
            "/admin/authentication/platformorganization/"
        )

    def test_list_display(self):
        self.assertEqual(
            self.platform_org_admin.list_display,
            (
                "name",
                "parent_organization",
                "org_category",
                "abbreviation",
                "child_organizations",
            ),
        )

    def test_list_editable(self):
        self.assertEqual(
            self.platform_org_admin.list_editable,
            ("parent_organization", "org_category", "abbreviation"),
        )

    def test_inlines(self):
        self.assertIn(PlatformOrganizationsInline, self.platform_org_admin.inlines)
        self.assertIn(
            PlatformOrganizationMembersInline, self.platform_org_admin.inlines
        )


class TestOrganizationCategoryAdmin(TestCase):
    def setUp(self):
        self.site = AdminSite()
        self.org_category_admin = OrganizationCategoryAdmin(
            OrganizationCategory, self.site
        )
        self.request = RequestFactory().get(
            "/admin/authentication/organizationcategory/"
        )

    def test_list_display(self):
        self.assertEqual(self.org_category_admin.list_display, ("name", "description"))



import logging
from typing import Dict, Type, TypedDict

from django.core.exceptions import ObjectDoesNotExist
from django.db import models
from rest_framework import exceptions, serializers

from smoothglue.core.field_trie import ModifiedTrie
from smoothglue.core.utils import check_kwargs

logger = logging.getLogger(__name__)


class NestedFieldTypes(TypedDict):
    serializer: Type[serializers.BaseSerializer]
    model: Type[models.Model]


class DynamicModelSerializer(serializers.ModelSerializer):
    """
    A serializer base class that allows removing fields from the response
    payload via the `fields` parameter in the query string.
    """

    # Limit number of fields if they are provided
    def __init__(self, *args, **kwargs):
        # Instantiate the superclass normally
        super().__init__(*args, **kwargs)

        # Not every serializer has a request associated
        if "request" in self.context:
            fields = self.context["request"].query_params.get("fields")
            if fields:
                trie = ModifiedTrie()
                requested_fields = fields.split(",")
                for rf in requested_fields:
                    trie.insert(rf)
                trie.filter_requested_fields(self.fields)

    def validate(self, attrs):
        request = self.context.get("request")
        if (
            self.instance is None
            and "id" in attrs
            and (self.partial or (request and request.method == "PATCH"))
        ):
            self.instance = self.Meta.model.objects.get(id=attrs["id"])
        return super().validate(attrs)

    def save(self, **kwargs):
        instance = super().save(**kwargs)
        return instance


class NestedObjectModelSerializer(DynamicModelSerializer):
    """
    Serializer base class that facilitates HTTP methods POST, PUT, and PATCH for objects
    nested inside a parent's serializer.

    i.e.
        `Asset` -> use `DynamicModelSerializer`,
        `Position` (child of `Asset`)  -> use `NestedObjectModelSerializer`

        # parent serializer class
        class AssetSerializer(DynamicModelSerializer):
            positions = PositionNestedSerializer(...)  # <- nested child here
            ...

            # nested config
            parent_id_field = "asset_id"  # <- tells nested children what to look for
            # lets parent serializer methods find paths to all create/updates
            nested_field_config = {
                "positions": {"serializer": PositionSerializer, "model": Position},
                ...
            }
            ...

        # technically this acts as a parent class with the functionality to also be nested
        class PositionSerializer(NestedObjectModelSerializer):
            asset_id = serializers.PrimaryKeyRelatedField(
                            source="asset",
                            queryset=Asset.objects.all()
                        )
            ...

        # the child/nested version of the above "parent" class version
        class PositionNestedSerializer(PositionSerializer):
            # parent id `asset_id` NOT required to allow for object creation/assignment
            asset_id = serializers.PrimaryKeyRelatedField(
                source="asset", queryset=Asset.objects.all(), required=False
            )
    """

    parent_id_field = None
    nested_field_config = {}

    def __init__(self, instance=None, data=serializers.empty, **kwargs):
        super().__init__(instance, data, **kwargs)
        self.nested_field_config = self.get_nested_field_config()
        self.parent_id_field = self.get_parent_id_field()

    def get_source_field_name(self, field_name: str):
        # pylint: disable=E1101
        source_field_name = self._declared_fields.get(field_name).source
        if not source_field_name:
            source_field_name = field_name
        return source_field_name

    def get_parent_id_field(self):
        assert self.parent_id_field is not None, (
            "'NestedObjectModelSerializer' should either include a `parent_id_field` "
            "attribute, or override the `get_parent_id_field()` method."
        )
        return self.parent_id_field

    def get_nested_field_config(self):
        """
        Get a `nested_field_config` which should have the following structure:
            nested_field_config = {
                "<field_name>": {
                    "serializer": <ModelSerializer>,
                    "model": <Model>,
                    "partial": <bool>
                },
                ...
            }

        Example:
            nested_field_config = {
                "positions": {"serializer": PositionSerializer, "model": Position},
            }
        """
        assert self.nested_field_config != {}, (
            "'NestedObjectModelSerializer' should either include a `nested_field_config` "
            "attribute, or override the `get_nested_field_config()` method."
        )
        return self.nested_field_config

    # pylint: disable=E1101
    def create(self, validated_data):
        obj_data = self.get_obj_data(self.nested_field_config, validated_data)
        parent_obj = self.Meta.model.objects.create(**validated_data)
        # Create/Update nested objects
        for field, d in self.nested_field_config.items():
            partial = d.get("partial", self.partial)
            self.iterate_nested_data(
                obj_data=obj_data,
                parent_instance=parent_obj,
                model=d["model"],
                field=field,
                serializer=d["serializer"],
                method="create",
                partial=partial,
            )

        return parent_obj

    def update(self, instance, validated_data):
        obj_data = self.get_obj_data(self.nested_field_config, validated_data)
        # Update/Create nested objects
        for field, d in self.nested_field_config.items():
            model = d["model"]
            source_field_name = self.get_source_field_name(field)
            many_to_many = instance._meta.get_field(source_field_name).many_to_many
            one_to_one = instance._meta.get_field(source_field_name).one_to_one
            partial = d.get("partial", self.partial)
            updated_ids = self.iterate_nested_data(
                obj_data=obj_data,
                parent_instance=instance,
                model=d["model"],
                field=field,
                serializer=d["serializer"],
                method="update",
                partial=partial,
            )
            # In case of PUT, clean up removed references
            if not partial and not many_to_many and not one_to_one:
                to_delete_objects = (
                    model.objects.filter(**{self.parent_id_field: instance.id})
                    .exclude(pk__in=updated_ids)
                    .all()
                )
                for obj in to_delete_objects:
                    obj.delete()

        return super().update(instance, validated_data)

    def iterate_nested_data(self, **kwargs):
        # Check args and unpack
        check_kwargs(
            ["obj_data", "parent_instance", "model", "field", "serializer", "method"],
            kwargs,
        )
        obj_data = kwargs["obj_data"]
        parent_instance = kwargs["parent_instance"]
        field = kwargs["field"]
        method = kwargs["method"]
        partial = kwargs["partial"]
        assert method in ["update", "create"]

        updated_ids = []
        source_field_name = self.get_source_field_name(field)
        many_to_many = parent_instance._meta.get_field(source_field_name).many_to_many
        if field in obj_data:
            for data in obj_data[field]:
                related_name = None
                relationship_ids = None
                if many_to_many:
                    related_name = parent_instance._meta.get_field(
                        field
                    ).remote_field.attname
                    relationship_ids = data.pop(related_name)
                else:
                    data[self.parent_id_field] = parent_instance.id
                child_instance = handle_nested_objects(
                    data, kwargs["model"], kwargs["serializer"], partial
                )
                updated_ids.append(child_instance.id)
                handle_many_to_many_objects(
                    child_instance=child_instance,
                    related_name=related_name,
                    relationship_ids=relationship_ids,
                    updated_ids=updated_ids,
                    **kwargs,
                )
        return updated_ids

    def get_obj_data(
        self, nested_field_objects: Dict[str, NestedFieldTypes], validated_data
    ):
        """
        Pull out the required fields (i.e. nested objects field) and return as dict
        """
        obj_data = {}
        for field in nested_field_objects:
            source_field_name = self.get_source_field_name(field)
            if source_field_name in validated_data:
                obj_data[field] = validated_data.pop(source_field_name)
        return obj_data


def handle_many_to_many_objects(**kwargs):
    # Check args and unpack
    check_kwargs(
        [
            "child_instance",
            "parent_instance",
            "related_name",
            "relationship_ids",
            "field",
            "updated_ids",
            "partial",
            "method",
        ],
        kwargs,
    )
    related_name = kwargs["related_name"]
    relationship_ids = kwargs["relationship_ids"]
    parent_instance = kwargs["parent_instance"]
    child_instance = kwargs["child_instance"]
    field = kwargs["field"]
    method = kwargs["method"]
    assert method in ["update", "create"]

    if related_name and relationship_ids:
        if method == "update":
            getattr(child_instance, related_name).set(relationship_ids)
            if kwargs["partial"]:
                getattr(parent_instance, field).add(child_instance)
            else:
                getattr(parent_instance, field).set(kwargs["updated_ids"])
        else:
            getattr(child_instance, related_name).set(relationship_ids)
            getattr(parent_instance, field).add(child_instance)


def handle_nested_objects(data, model, serializer, partial=False):
    """
    Takes a nested object and performs update or create dependent on 'id' being present
    """
    data_id = data.get("id")
    if data_id:
        try:
            obj_instance = model.objects.get(id=data_id)
        except ObjectDoesNotExist as ex:
            raise exceptions.NotFound(
                detail=f"{model.__name__} with id: {data_id} not found."
            ) from ex
        obj_instance = serializer(partial=partial).update(obj_instance, data)
    elif not partial:  # i.e. PUT
        obj_instance = serializer().create(data)
    else:
        raise serializers.ValidationError(
            f"'id' required for nested update to {model.__name__}.", code="bad_request"
        )
    return obj_instance



import copy
import logging
from typing import Union

from django.db.models import Model
from django.utils import timezone
from rest_framework.request import Request

from smoothglue.tracker.models import APIChangeLog

logger = logging.getLogger(__name__)


# pylint: disable=W0212
def create_change_log(request: Request, instance: Union[Model, None] = None) -> None:
    """
    Creates an APIChangeLog from a request
    """
    try:
        data = copy.deepcopy(request.data)
        if isinstance(data, dict) and instance:
            data.update(id=str(instance.id))

        APIChangeLog.objects.create(
            username=str(request.user)[:200],
            full_path=request._request.path[:200],
            method=request._request.method,
            timestamp=timezone.now(),
            data=data,
            params=request.query_params.dict(),
        )

    # Currently unsure what exceptions could occur, so using broad exception catch
    # pylint: disable=W0703
    except Exception as ex:
        logger.error(
            "Failed to log exception to db - %s: %s",
            type(ex).__name__,
            str(ex),
            exc_info=True,
        )



from smoothglue.core.serializers.dynamic import DynamicModelSerializer
from smoothglue.tracker.models import APIChangeLog, AppErrorLog


class APIChangeLogSerializer(DynamicModelSerializer):
    class Meta:
        model = APIChangeLog
        fields = [
            "id",
            "username",
            "full_path",
            "method",
            "timestamp",
            "data",
            "params",
        ]


class AppErrorLogSerializer(DynamicModelSerializer):
    def create(self, validated_data):
        """
        Auto assign the current request user as the AppErrorLog's user.
        Added request HTTP_HEADER to the client_data
        """
        request = self.context.get("request")
        if request:
            return AppErrorLog.objects.create(
                user=request.user,
                level=validated_data.get("level"),
                message=validated_data.get("message"),
                stack_trace=validated_data.get("stack_trace"),
                client_data={
                    key: request.META[key]
                    for key in request.META
                    if key.startswith("HTTP_")
                },
            )
        return AppErrorLog.objects.create(
            user=None,
            level=validated_data.get("level"),
            message=validated_data.get("message"),
            stack_trace=validated_data.get("stack_trace"),
            client_data={},
        )

    class Meta:
        model = AppErrorLog
        fields = [
            "level",
            "message",
            "stack_trace",
            "user",
            "client_data",
        ]



from django.urls import include, path
from rest_framework import routers

from .views import (
    OrganizationCategoryViewSet,
    OrganizationMemberViewSet,
    PlatformOrganizationViewSet,
    PlatformUserViewSet,
    get_active_user,
    what_are_my_headers,
)

# Router setup
ROUTER = routers.DefaultRouter()
ROUTER.register(r"users", PlatformUserViewSet, basename="users")
ROUTER.register(r"organizations", PlatformOrganizationViewSet)
ROUTER.register(r"org-members", OrganizationMemberViewSet)
ROUTER.register(r"org-categories", OrganizationCategoryViewSet)

# URL Patterns
urlpatterns = [
    path("active_user/", get_active_user, name="get_active_user"),
    path("headers/", what_are_my_headers, name="what_are_my_headers"),
    path("", include(ROUTER.urls)),
]



#!/usr/bin/env python
import sys

from django.core.management import call_command

from smoothglue.core.scripts.boot_django import boot_django

# call the django setup routine
boot_django()

call_command("loaddata", sys.argv[1])



from django.contrib.auth import get_user_model
from django.urls import reverse
from django.utils import timezone
from rest_framework import status
from rest_framework.test import APITestCase

from smoothglue.tracker.models import APIChangeLog, AppErrorLog

User = get_user_model()

MOCK_RESOURCE_PATH = "/api/resource1"


class AppErrorLogViewSetTests(APITestCase):
    def setUp(self):
        self.user = User.objects.create_user(
            username="testuser", password="password123"
        )
        self.url = reverse("apperrorlogger-list")

    def test_create_error_log_authenticated(self):
        """
        Ensure authenticated users can POST error logs.
        Verify user and client_data (headers) are saved via serializer context.
        """
        self.client.force_authenticate(user=self.user)
        valid_payload = {
            "level": "ERROR",
            "message": "Something went wrong on the client.",
            "stack_trace": "Traceback details...",
        }
        response = self.client.post(
            self.url,
            data=valid_payload,
            format="json",
            HTTP_USER_AGENT="TestAgent/1.0",
            HTTP_X_CUSTOM_INFO="TestData",
        )

        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.assertEqual(AppErrorLog.objects.count(), 1)
        log_entry = AppErrorLog.objects.first()

        self.assertEqual(log_entry.level, valid_payload["level"])
        self.assertEqual(log_entry.message, valid_payload["message"])
        self.assertEqual(log_entry.stack_trace, valid_payload["stack_trace"])
        self.assertEqual(log_entry.user, self.user)

    def test_create_error_log_invalid_data(self):
        """
        Ensure POST requests with invalid data (e.g., missing required fields) fail.
        """
        self.client.force_authenticate(user=self.user)
        invalid_payload = {"stack_trace": "Incomplete data."}
        response = self.client.post(self.url, data=invalid_payload, format="json")

        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(AppErrorLog.objects.count(), 0)
        self.assertIn("level", response.data)
        self.assertIn("message", response.data)

    def test_disallowed_methods_on_error_log(self):
        """
        Ensure GETs are not allowed for AppErrorLogViewSet.
        """
        self.client.force_authenticate(user=self.user)

        AppErrorLog.objects.create(level="INFO", message="temp", user=self.user)

        # Test list methods
        response_get_list = self.client.get(self.url)
        self.assertEqual(
            response_get_list.status_code, status.HTTP_405_METHOD_NOT_ALLOWED
        )


class APIChangeLogViewSetTests(APITestCase):
    def setUp(self):
        self.user1 = User.objects.create_user(
            username="user1", password="password123", email="user1@example.com"
        )
        self.user2 = User.objects.create_user(
            username="user2", password="password456", email="user2@example.com"
        )

        self.log1 = APIChangeLog.objects.create(
            username="user1",
            full_path=MOCK_RESOURCE_PATH,
            method="GET",
            timestamp=timezone.now(),
        )
        self.log2 = APIChangeLog.objects.create(
            username="user1",
            full_path="/api/resource2",
            method="POST",
            timestamp=timezone.now(),
        )
        self.log3 = APIChangeLog.objects.create(
            username="user2",
            full_path=MOCK_RESOURCE_PATH,
            method="GET",
            timestamp=timezone.now(),
        )

        self.list_url = reverse("apichangelog-list")
        self.detail_url_log1 = reverse(
            "apichangelog-detail", kwargs={"pk": self.log1.pk}
        )
        self.detail_url_nonexistent = reverse(
            "apichangelog-detail", kwargs={"pk": 9999}
        )
        self.client.force_authenticate(user=self.user1)

    def test_list_change_logs(self):
        """
        Ensure GET request to list endpoint returns logs.
        """
        response = self.client.get(self.list_url)
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertEqual(len(response.data), 3)
        self.assertIsInstance(response.data, list)

    def test_filter_change_logs_by_username(self):
        """
        Test filtering the list by the 'username' query parameter.
        """
        response = self.client.get(self.list_url, {"username": "user1"})
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(len(response.data), 2)  # Only log1 and log2 are for user1
        self.assertTrue(all(item["username"] == "user1" for item in response.data))

        response_user2 = self.client.get(self.list_url, {"username": "user2"})
        self.assertEqual(response_user2.status_code, status.HTTP_200_OK)
        self.assertEqual(len(response_user2.data), 1)  # Only log3 is for user2
        self.assertEqual(response_user2.data[0]["id"], str(self.log3.id))

        response_nouser = self.client.get(self.list_url, {"username": "nouser"})
        self.assertEqual(response_nouser.status_code, status.HTTP_200_OK)
        self.assertEqual(len(response_nouser.data), 0)

    def test_filter_change_logs_by_full_path(self):
        """
        Test filtering the list by the 'full_path' query parameter.
        """
        response = self.client.get(self.list_url, {"full_path": MOCK_RESOURCE_PATH})
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(len(response.data), 2)  # log1 and log3 match path
        results_ids = {item["id"] for item in response.data}
        self.assertIn(str(self.log1.id), results_ids)
        self.assertIn(str(self.log3.id), results_ids)

        response_path2 = self.client.get(self.list_url, {"full_path": "/api/resource2"})
        self.assertEqual(response_path2.status_code, status.HTTP_200_OK)
        self.assertEqual(len(response_path2.data), 1)  # Only log2 matches
        self.assertEqual(response_path2.data[0]["id"], str(self.log2.id))

    def test_retrieve_change_log(self):
        """
        Ensure GET request to detail endpoint retrieves a specific log.
        """
        response = self.client.get(self.detail_url_log1)
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data["id"], str(self.log1.id))
        self.assertEqual(response.data["username"], self.log1.username)
        self.assertEqual(response.data["full_path"], self.log1.full_path)

    def test_retrieve_change_log_not_found(self):
        """
        Ensure GET request for a non-existent ID returns 404 Not Found.
        """
        response = self.client.get(self.detail_url_nonexistent)
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_disallowed_methods_on_change_log(self):
        """
        Ensure POST, PUT, PATCH, DELETE are not allowed for APIChangeLogViewSet.
        """
        response_post = self.client.post(self.list_url, data={}, format="json")
        self.assertEqual(response_post.status_code, status.HTTP_405_METHOD_NOT_ALLOWED)

        detail_url = self.detail_url_log1
        response_put = self.client.put(detail_url, data={}, format="json")
        self.assertEqual(response_put.status_code, status.HTTP_405_METHOD_NOT_ALLOWED)

        response_patch = self.client.patch(detail_url, data={}, format="json")
        self.assertEqual(response_patch.status_code, status.HTTP_405_METHOD_NOT_ALLOWED)

        response_delete = self.client.delete(detail_url)
        self.assertEqual(
            response_delete.status_code, status.HTTP_405_METHOD_NOT_ALLOWED
        )



"""Tests for app_info serializers to improve coverage."""

import pytest

from smoothglue.core.models import AppInfo, Classification
from smoothglue.core.serializers.app_info import (
    AppInfoSerializer,
    ClassificationSerializer,
)


class TestClassificationSerializer:
    """Test ClassificationSerializer create and update methods."""

    def test_create_classification(self):
        """Test creating a new Classification instance."""
        serializer = ClassificationSerializer()
        validated_data = {"level": "unclassified", "message": "Test message"}

        classification = serializer.create(validated_data)

        assert isinstance(classification, Classification)
        assert classification.level == "unclassified"
        assert classification.message == "Test message"

    def test_update_classification(self):
        """Test updating an existing Classification instance."""
        serializer = ClassificationSerializer()
        instance = Classification(level="secret", message="Old message")
        validated_data = {"level": "unclassified", "message": "New message"}

        updated = serializer.update(instance, validated_data)

        assert updated.level == "unclassified"
        assert updated.message == "New message"

    def test_update_classification_partial(self):
        """Test partial update of Classification instance."""
        serializer = ClassificationSerializer()
        instance = Classification(level="secret", message="Old message")
        validated_data = {"level": "unclassified"}

        updated = serializer.update(instance, validated_data)

        assert updated.level == "unclassified"
        assert updated.message == "Old message"  # Should remain unchanged


class TestAppInfoSerializer:
    """Test AppInfoSerializer create and update methods."""

    def test_create_app_info(self):
        """Test creating a new AppInfo instance."""
        serializer = AppInfoSerializer()
        classification = Classification(level="unclassified", message="Test")
        validated_data = {"classification": classification}

        app_info = serializer.create(validated_data)

        assert isinstance(app_info, AppInfo)
        assert app_info.classification == classification

    def test_update_app_info(self):
        """Test updating an existing AppInfo instance."""
        serializer = AppInfoSerializer()
        old_classification = Classification(level="secret", message="Old")
        new_classification = Classification(level="unclassified", message="New")
        instance = AppInfo(classification=old_classification)
        validated_data = {"classification": new_classification}

        updated = serializer.update(instance, validated_data)

        assert updated.classification == new_classification

    def test_update_app_info_no_classification(self):
        """Test updating AppInfo without providing classification."""
        serializer = AppInfoSerializer()
        classification = Classification(level="secret", message="Original")
        instance = AppInfo(classification=classification)
        validated_data = {}

        updated = serializer.update(instance, validated_data)

        assert updated.classification == classification  # Should remain unchanged



from unittest.mock import patch

import jwt
from django.contrib.auth import authenticate
from django.core.exceptions import ValidationError
from django.http import HttpRequest
from django.test import TestCase
from django.test.utils import override_settings

from smoothglue.authentication.auth import (
    PlatformAuthentication,
    decode_jwt,
    get_unknown_user,
)
from smoothglue.authentication.defaults import ROOT_ADMIN_USERNAME


class TestDecodeJWT(TestCase):
    def setUp(self):
        self.request = HttpRequest()

    @patch("jwt.decode")
    def test_decode_jwt_with_authorization_header(self, mock_jwt_decode):
        mock_jwt_decode.return_value = {"some": "data"}
        self.request.META["HTTP_AUTHORIZATION"] = "Bearer testtoken"
        result = decode_jwt(self.request)
        self.assertEqual(result, {"some": "data"})
        mock_jwt_decode.assert_called_once_with(
            "testtoken", options={"verify_signature": False}
        )

    @patch("jwt.decode")
    def test_decode_jwt_with_jwt_header(self, mock_jwt_decode):
        mock_jwt_decode.return_value = {"some": "data"}
        self.request.META["HTTP_JWT"] = "testtoken"
        result = decode_jwt(self.request)
        self.assertEqual(result, {"some": "data"})
        mock_jwt_decode.assert_called_once_with(
            "testtoken", options={"verify_signature": False}
        )

    def test_decode_jwt_no_token(self):
        with self.assertRaises(ValueError) as context:
            decode_jwt(self.request)
        self.assertEqual(str(context.exception), "No JWT provided in header.")

    @patch("jwt.decode", side_effect=jwt.DecodeError("Invalid token"))
    def test_decode_jwt_invalid_token(self, mock_jwt_decode):
        self.request.META["HTTP_AUTHORIZATION"] = "Bearer invalidtoken"
        with self.assertRaises(ValueError) as context:
            decode_jwt(self.request)
        self.assertEqual(str(context.exception), "Invalid JWT token: Invalid token")


class TestGetUnknownUser(TestCase):
    @override_settings(SINGLE_USER_DEFAULT_EMAIL="default@example.com")
    def test_get_unknown_user(self):
        user = get_unknown_user()
        self.assertEqual(user.email, "default@example.com")
        self.assertEqual(user.username, "UnknownUser")
        self.assertTrue(user.is_superuser)


class TestPlatformAuthentication(TestCase):
    def setUp(self):
        self.auth = PlatformAuthentication()
        self.request = HttpRequest()

    @override_settings(ENABLE_SINGLE_USER_MODE=True)
    @patch(
        "smoothglue.authentication.auth.get_unknown_user",
        return_value={"id": 1, "name": "foo"},
    )
    def test_authenticate_single_user_mode(self, mock_get_unknown_user):
        user, _ = self.auth.authenticate(self.request)
        self.assertEqual(user, {"id": 1, "name": "foo"})

    @patch("smoothglue.authentication.auth.decode_jwt")
    @patch("django.contrib.auth.get_user_model")
    def test_authenticate_with_valid_jwt(self, mock_get_user_model, mock_decode_jwt):
        mock_decode_jwt.return_value = {
            "email": "test@example.com",
            "preferred_username": "testuser",
            "given_name": "Test",
            "family_name": "User",
            "rank": "N/A",
            "usercertificate": "N/A",
            "affiliation": "N/A",
        }
        self.request.META["HTTP_AUTHORIZATION"] = "Bearer validtoken"
        user, _ = self.auth.authenticate(self.request)
        self.assertEqual(user.email, "test@example.com")
        self.assertEqual(user.username, "testuser")
        self.assertFalse(user.is_superuser)

    @patch("smoothglue.authentication.auth.decode_jwt")
    def test_authenticate_with_invalid_jwt(self, mock_decode_jwt):
        mock_decode_jwt.side_effect = ValueError("Invalid JWT token")
        self.request.META["HTTP_AUTHORIZATION"] = "Bearer invalidtoken"
        with self.assertRaises(ValueError) as context:
            self.auth.authenticate(self.request)
        self.assertEqual(str(context.exception), "Invalid JWT token")

    def test_is_valid_email(self):
        self.assertTrue(self.auth.is_valid_email("valid@example.com"))

    def test_is_invalid_email(self):
        self.assertFalse(self.auth.is_valid_email("invalid"))

    @patch("django.contrib.auth.validators.UnicodeUsernameValidator")
    def test_is_valid_username(self, mock_unicode_username_validator):
        validator_instance = mock_unicode_username_validator.return_value
        self.assertTrue(self.auth.is_valid_username("validusername"))
        validator_instance.assert_called_once_with("validusername")

    @patch("django.contrib.auth.validators.UnicodeUsernameValidator")
    def test_is_invalid_username(self, mock_unicode_username_validator):
        validator_instance = mock_unicode_username_validator.return_value
        validator_instance.side_effect = ValidationError("Invalid username")
        self.assertFalse(self.auth.is_valid_username("invalidusername"))
        validator_instance.assert_called_once_with("invalidusername")


class TestRootAdminBackend(TestCase):
    @override_settings(
        AUTHENTICATION_BACKENDS=["smoothglue.authentication.auth.RootAdminBackend"]
    )
    def test_root_admin(self):
        test_password = "test-password"
        with override_settings(ROOT_ADMIN_PASSWORD=test_password):
            user = authenticate(username=ROOT_ADMIN_USERNAME, password=test_password)
            self.assertEqual(user.username, ROOT_ADMIN_USERNAME)
            self.assertTrue(user.is_staff)
            self.assertTrue(user.is_superuser)


class TestPlatformAuthenticationMiddleware(TestCase):
    def setUp(self):
        self.test_user = get_user_model().objects.create(
            username="test", email="test@localhost"
        )

    def test_middleware(self):
        with patch(
            "smoothglue.authentication.auth.PlatformAuthentication.authenticate",
            return_value=(self.test_user, None),
        ):
            response = self.client.get(
                "/admin/", headers={"Authorization": "Bearer test_token"}
            )
            self.assertEqual(response.wsgi_request.user, self.test_user)



from drf_spectacular.extensions import OpenApiSerializerExtension


class CustomDataSchema(OpenApiSerializerExtension):

    target_class = "smoothglue.authentication.serializers.PlatformUserSerializer"

    def map_serializer(self, auto_schema, direction):
        # pylint: disable=import-outside-toplevel
        # pylint: disable=no-name-in-module
        from smoothglue.authentication.serializers import UserDataSerializer

        # pylint: disable=inherit-non-class
        class Fixed(self.target_class):

            data = UserDataSerializer()

        return auto_schema._map_serializer(Fixed, direction)



from smoothglue.core.views._async_views import AsyncListModelViewSet
from smoothglue.core.views._views import (
    CustomDataSchema,
    ListModelViewSet,
    get_app_info,
)

__all__ = [
    "ListModelViewSet",
    "CustomDataSchema",
    "get_app_info",
    "AsyncListModelViewSet",
]



import re
import sys

from pylint import run_pylint

from smoothglue.core.scripts.boot_django import boot_django


def _run_pylint():
    boot_django()
    sys.argv[0] = re.sub(r"(-script\.pyw|\.exe)?$", "", sys.argv[0])
    sys.exit(run_pylint())


if __name__ == "__main__":
    _run_pylint()



from unittest.mock import MagicMock, patch

from django.contrib.auth import get_user_model
from django.test import TestCase
from django.utils import timezone

from smoothglue.tracker.models import APIChangeLog, AppErrorLog
from smoothglue.tracker.serializers import APIChangeLogSerializer, AppErrorLogSerializer

User = get_user_model()

FAKE_STACK_TRACE = "Trace..."


class APIChangeLogSerializerTest(TestCase):
    def setUp(self):
        self.user_obj = User.objects.create_user(
            username="testlogger", password="password123"
        )
        self.log_timestamp = timezone.now()
        self.api_log_instance = APIChangeLog.objects.create(
            username=self.user_obj.username,
            full_path="/api/test/endpoint",
            method="GET",
            timestamp=self.log_timestamp,
            data={"request_data": "sample"},
            params={"query_param": "value"},
        )

    def test_api_change_log_serialization(self):
        """
        Test that APIChangeLogSerializer correctly serializes an APIChangeLog instance.
        """
        serializer = APIChangeLogSerializer(instance=self.api_log_instance)
        data = serializer.data

        self.assertEqual(data["id"], str(self.api_log_instance.id))
        self.assertEqual(data["username"], self.api_log_instance.username)
        self.assertEqual(data["full_path"], "/api/test/endpoint")
        self.assertEqual(data["method"], "GET")
        self.assertEqual(data["data"], {"request_data": "sample"})
        self.assertEqual(data["params"], {"query_param": "value"})
        # checking that timestamp is in data due to potential
        # microsecond differences related to instantiation and test run
        self.assertTrue("timestamp" in data)
        self.assertEqual(
            data["timestamp"], self.log_timestamp.isoformat().replace("+00:00", "Z")
        )

        expected_keys = {
            "id",
            "username",
            "full_path",
            "method",
            "timestamp",
            "data",
            "params",
        }
        self.assertEqual(set(data.keys()), expected_keys)

    def test_api_change_log_deserialization_validation(self):
        """
        Test basic validation for APIChangeLogSerializer during deserialization.
        """
        valid_data = {
            "username": "newuser",
            "full_path": "/api/v2/resource",
            "method": "POST",
            "timestamp": timezone.now().isoformat(),
            "data": {"key": "payload"},
            "params": {"filter": "active"},
        }
        serializer = APIChangeLogSerializer(data=valid_data)
        self.assertTrue(serializer.is_valid())


class AppErrorLogSerializerTest(TestCase):
    def setUp(self):
        self.user = User.objects.create_user(
            username="errorreporter", password="password123"
        )
        self.app_error_instance = AppErrorLog.objects.create(
            user=self.user,
            level="ERROR",
            message="Initial error for serialization test.",
            stack_trace="Traceback...",
            client_data={"HTTP_USER_AGENT": "TestClient"},
        )
        # Mock request object for context
        self.mock_request = MagicMock()
        self.mock_request.user = self.user
        self.mock_request.META = {
            "HTTP_USER_AGENT": "Chrome (Test)",
            "HTTP_X_CUSTOM_HEADER": "CustomValue",
            "CONTENT_TYPE": "application/json",  # Should not be included in client_data
            "REMOTE_ADDR": "foo",  # Should not be included
        }
        self.serializer_context = {"request": self.mock_request}

    def test_app_error_log_serialization(self):
        """
        Test that AppErrorLogSerializer correctly serializes an AppErrorLog instance.
        """
        serializer = AppErrorLogSerializer(instance=self.app_error_instance)
        data = serializer.data

        self.assertEqual(data["level"], "ERROR")
        self.assertEqual(data["message"], "Initial error for serialization test.")
        self.assertEqual(data["stack_trace"], "Traceback...")
        self.assertEqual(data["user"], self.user.pk)
        self.assertEqual(data["client_data"], "{'HTTP_USER_AGENT': 'TestClient'}")

        expected_keys = {"level", "message", "stack_trace", "user", "client_data"}
        self.assertEqual(set(data.keys()), expected_keys)

    def test_app_error_log_create_with_request_context(self):
        """
        Test the create method of AppErrorLogSerializer with a request in context.
        """
        valid_data = {
            "level": "WARNING",
            "message": "A new warning occurred.",
            "stack_trace": "Detailed traceback here.",
        }
        serializer = AppErrorLogSerializer(
            data=valid_data, context=self.serializer_context
        )
        self.assertTrue(serializer.is_valid())

    def test_app_error_log_create_without_request_context(self):
        """
        Test create method when 'request' is not in the serializer context.
        """
        valid_data = {
            "level": "ERROR",
            "message": "Informational message.",
            "stack_trace": FAKE_STACK_TRACE,
        }
        # No 'request' in context
        serializer = AppErrorLogSerializer(data=valid_data, context={})
        self.assertTrue(serializer.is_valid())

        instance = serializer.save()
        self.assertEqual(instance.level, "ERROR")
        self.assertEqual(instance.message, "Informational message.")
        self.assertEqual(instance.stack_trace, FAKE_STACK_TRACE)
        self.assertEqual(instance.client_data, {})
        self.assertIsNone(instance.user)

    def test_app_error_log_create_with_anonymous_user(self):
        """
        Test create method when request.user is None.
        """
        self.mock_request.user = (
            None  # Simulate AnonymousUser or request.user being None
        )
        valid_data = {
            "level": "DEBUG",
            "message": "Debug message from anonymous.",
            "stack_trace": FAKE_STACK_TRACE,
        }
        serializer = AppErrorLogSerializer(
            data=valid_data, context=self.serializer_context
        )
        self.assertTrue(serializer.is_valid())

        with patch(
            "smoothglue.tracker.models.AppErrorLog.objects.create"
        ) as mock_db_create:
            created_log_mock = MagicMock(
                spec=AppErrorLog, user=None
            )  # Simulate instance with user=None
            mock_db_create.return_value = created_log_mock

            instance = serializer.save()

            mock_db_create.assert_called_once()
            call_args = mock_db_create.call_args[1]
            self.assertIsNone(call_args["user"])
            self.assertEqual(instance, created_log_mock)

    def test_app_error_log_create_with_no_http_headers(self):
        """
        Test client_data is an empty dict.
        """
        self.mock_request.META = {"CONTENT_TYPE": "text/plain"}  # No HTTP headers
        valid_data = {
            "level": "CRITICAL",
            "message": "Critical issue, no headers.",
            "stack_trace": "Critical trace.",
        }
        serializer = AppErrorLogSerializer(
            data=valid_data, context=self.serializer_context
        )
        self.assertTrue(serializer.is_valid())

        with patch(
            "smoothglue.tracker.models.AppErrorLog.objects.create"
        ) as mock_db_create:
            created_log_mock = MagicMock(spec=AppErrorLog, client_data={})
            mock_db_create.return_value = created_log_mock

            serializer.save()
            call_args = mock_db_create.call_args[1]
            self.assertEqual(call_args["client_data"], {})



import logging

from django_filters.rest_framework import DjangoFilterBackend
from rest_framework import mixins, viewsets
from rest_framework.pagination import LimitOffsetPagination

from smoothglue.tracker.models import APIChangeLog
from smoothglue.tracker.serializers import APIChangeLogSerializer, AppErrorLogSerializer

logger = logging.getLogger(__name__)


class AppErrorLogViewSet(mixins.CreateModelMixin, viewsets.GenericViewSet):
    """
    This provides the client a last resort for notifying the backend that an exception happened.
    This silents all exceptions to ensure that the client side request doesn't fail when calling
    the fail safe endpoint.

    Only POST is allowed as viewing the exception data should be done within Django Admin

    """

    serializer_class = AppErrorLogSerializer


class APIChangeLogViewSet(viewsets.ReadOnlyModelViewSet):
    """
    API for retrieving API execution logs
    """

    filter_backends = [DjangoFilterBackend]
    filterset_fields = ["username", "full_path"]
    queryset = APIChangeLog.objects.all()
    serializer_class = APIChangeLogSerializer
    pagination_class = LimitOffsetPagination
